import os
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Any, List, Optional, Union
from elasticsearch import Elasticsearch


def generate_all_manifests(es_host: str,
                           collection_index: str,
                           manifest_root_path: str,
                           index_list_path: Optional = None):
    es = Elasticsearch(es_host)
    for product_index in get_all_product_indices(es, collection_index):
        if should_make_manifest(product_index.index, index_list_path):
            manifests: List[str] = generate_manifest(es, product_index)
            write_results(manifests, get_result_path(manifest_root_path, product_index))


@dataclass
class ProductIndex:
    index: str
    href_to_replace: str
    href_replacement: str

    @classmethod
    def from_collection_metadata(cls, metadata: Dict[str, Any]):
        index: str = metadata['_source']['id']
        href_to_replace: str = metadata['_source']['hiddenInformation']['dataAccess']["HTTP"]["src"]
        href_replacement: str = metadata['_source']['hiddenInformation']['dataAccess']["HTTP"]["dst"]
        return ProductIndex(index, href_to_replace, href_replacement)

    def transform_paths(self, paths: List[str]) -> List[str]:
        return [self._transform(path) for path in paths]

    def _transform(self, path: str):
        return path.replace(self.href_to_replace, self.href_replacement)

    @property
    def oscars_index(self) -> str:
        return f"oscars_{self.index}"

    @property
    def manifest_name(self) -> str:
        return f"manifest_{self.index}_latest.txt"


def should_make_manifest(product_index: str, index_list_path: Optional[str]) -> bool:
    if index_list_path is None:
        return True

    with open(index_list_path, 'r') as f:
        indices = f.read().splitlines()
        return product_index in indices


def generate_manifest(es, product_index):
    all_paths = get_product_paths(es, product_index.oscars_index)
    return product_index.transform_paths(all_paths)


def get_all_product_indices(es: Elasticsearch, collection_index: str) -> List[ProductIndex]:
    body: Dict[str, Any] = {
        "query": {
            "match_all": {}
        },
        "size": 1000
    }
    responses: Dict[str, Any] = es.search(index=collection_index, body=body)
    return [ProductIndex.from_collection_metadata(product_index_metadata)
            for product_index_metadata in responses['hits']['hits']]


def get_product_paths(es: Elasticsearch, index_name: str) -> List[str]:
    def get_path(hit: Dict[str, Any]) -> str:
        return hit['_source']['properties']['links']['data'][0]['href']

    scroll_time = "2m"
    response = es.search(
        index=index_name,
        body={
              "query": {
                "bool": {
                  "must": {
                      "term": {
                        "properties.status":"ARCHIVED"
                      }
                    }
                }
              }
            },
        scroll=scroll_time,
        size=1000
    )
    scroll_id = response['_scroll_id']
    paths = [get_path(hit) for hit in response['hits']['hits']]

    while response['hits']['hits'] and '_scroll_id' in response:
        response = es.scroll(
            scroll_id=scroll_id,
            scroll=scroll_time
        )
        paths.extend([get_path(hit) for hit in response['hits']['hits']])
        scroll_id = response['_scroll_id']
    return paths


def get_result_path(parent_directory: str, product_index: ProductIndex) -> Path:
    return Path(parent_directory) / f"{product_index.index}/{product_index.manifest_name}"


def write_results(results: List[str], filename: Union[str, Path]) -> None:
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, 'w+') as f:
        for href in results:
            f.write("%s\n" % href)
